from typing import Tuple
from uuid import uuid4


class MultipartEncoder:
    def __init__(self, fields=None, files=None):
        # name, value
        self.fields: List[bytes, bytes] = fields or []

        # field_name, file_name, content_type, body
        self.files: List[bytes, bytes, bytes, bytes] = files or []

    def encode(self) -> Tuple[str, bytes]:
        boundary = uuid4().hex
        _boundary = f"--{boundary}".encode()
        data = b""

        if self.fields:
            data += b"".join(
                b"%s\r\n"
                b'Content-Disposition: form-data; name="%s"\r\n'
                b"\r\n"
                b"%s\r\n\r\n" % (_boundary, name, value)
                for name, value in self.fields
            )

        if self.files:
            data += b"".join(
                b"%s\r\n"
                b'Content-Disposition: form-data; name="%s"; filename="%s"\r\n'
                b"Content-Type: %s\r\n"
                b"\r\n"
                b"%s\r\n\r\n" % (_boundary, field_name, file_name, content_type, body)
                for field_name, file_name, content_type, body in self.files
            )

        data += b"%s--\r\n\r\n" % _boundary

        return f"multipart/form-data; boundary={boundary}", data



from enum import Enum
from mimetypes import guess_type as guess_mime
from typing import List, Optional, Tuple


# method, endpoint, headers, params, body
EndpointCall = Tuple[str, str, Optional[dict], Optional[dict], Optional[bytes]]


class ParseMode(Enum):
    Markdown = 1
    HTML = 2
    MarkdownV2 = 3


class InputFile:
    def __init__(self, file_id: str, *, body: Optional[bytes] = None, content_type: Optional[str] = None):
        # file id, file name or url
        self.file_id = file_id

        self.body = body
        if body:
            self.content_type = (content_type or guess_mime(file_id)[0]).encode()  # type: ignore

    def update(self, file_id: str):
        self.file_id = file_id
        self.body = None

    def __call__(
        self,
        field_name: str,
        files: List[Tuple[bytes, bytes, bytes, bytes]],
        params: dict,
        *,
        attach: Optional[str] = None,
    ):
        if (body := self.body) :
            if attach:
                files.append((attach.encode(), self.file_id.encode(), self.content_type, body))
                params[field_name] = f"attach://{attach}"
            else:
                files.append((field_name.encode(), self.file_id.encode(), self.content_type, body))
        else:
            params[field_name] = self.file_id


class InlineKeyboardMarkup:
    __slots__ = ["serialized", "__current_row"]

    def __init__(self):
        row = []
        self.serialized = {"inline_keyboard": [row]}
        self.__current_row = row

    def add_row(self):
        row = []
        self.serialized["inline_keyboard"].append(row)
        self.__current_row = row
        return self

    def url(self, text: str, url: str):
        self.__current_row.append({"text": text, "url": url})
        return self

    def login_url(
        self,
        text: str,
        url: str,
        *,
        forward_text: str = None,
        bot_username: str = None,
        request_write_access: bool = False,
    ):
        login_url: dict = {"url": url}

        if forward_text:
            login_url["forward_text"] = forward_text

        if bot_username:
            login_url["bot_username"] = bot_username

        if request_write_access:
            login_url["request_write_access"] = True

        self.__current_row.append({"text": text, "login_url": login_url})
        return self

    def callback_data(self, text: str, callback_data: bytes):
        assert 0 < len(callback_data) < 64
        self.__current_row.append({"text": text, "callback_data": callback_data})
        return self

    def switch_inline_query(self, text: str, switch_inline_query: str):
        self.__current_row.append({"text": text, "switch_inline_query": switch_inline_query})
        return self

    def switch_inline_query_current_chat(self, text: str, switch_inline_query_current_chat: str):
        self.__current_row.append({"text": text, "switch_inline_query_current_chat": switch_inline_query_current_chat})
        return self

    def callback_game(self, text: str):
        assert len(self.serialized["inline_keyboard"]) == 1 and len(self.__current_row) == 0
        self.__current_row.append({"text": text, "callback_game": {}})
        return self

    def pay(self, text: str):
        assert len(self.serialized["inline_keyboard"]) == 1 and len(self.__current_row) == 0
        self.__current_row.append({"text": text, "pay": True})
        return self


class ReplyKeyboardMarkup:
    __slots__ = ["serialized", "__current_row"]

    def __init__(self):
        row = []
        self.serialized = {"keyboard": [row]}
        self.__current_row = row

    def add_row(self):
        row = []
        self.serialized["keyboard"].append(row)
        self.__current_row = row
        return self

    def add_button(self, text, *, request_contact=False, request_location=False):
        button = {"text": text}

        if request_contact:
            button["request_contact"] = True

        if request_location:
            button["request_location"] = True

        self.__current_row.append(button)
        return self

    def resize_keyboard(self, resize_keyboard: bool = True):
        self.serialized["resize_keyboard"] = resize_keyboard
        return self

    def one_time_keyboard(self, one_time_keyboard: bool = True):
        self.serialized["one_time_keyboard"] = one_time_keyboard
        return self

    def selective(self, selective: bool = True):
        self.serialized["selective"] = selective
        return self


class ReplyKeyboardRemove:
    __slots__ = ["serialized"]

    def __init__(self):
        self.serialized = {"remove_keyboard": True}

    def selective(self, selective: bool = True):
        self.serialized["selective"] = selective
        return self


class ForceReply:
    __slots__ = ["serialized"]

    def __init__(self):
        self.serialized = {"force_reply": True}

    def selective(self, selective: bool = True):
        self.serialized["selective"] = selective
        return self



from os.path import dirname, realpath
from re import compile as re_compile

from lxml import etree


class IndexedTuple(tuple):
    def __hash__(self):
        return hash(self[0])


class ApiParser:
    __slots__ = ["__type_split_regex", "__primitive_types", "__extract_text", "__extract_all_text"]

    def __init__(self):
        self.__type_split_regex = re_compile(" or ")

        # primitive types
        self.__primitive_types = {"Float", "Integer", "String", "Array", "Boolean"}

        # function to extract the text from the node
        self.__extract_text = etree.XPath("string()")

        # function to extract the text from the node and the descendant nodes
        extract_all_text_list = etree.XPath("descendant-or-self::text()")
        self.__extract_all_text = lambda node: "".join(extract_all_text_list(node))

    def __extract_type_list(self, cell, extract_all_text):
        return tuple(
            # normalize "Float number" to "Float"
            ("Float",) if _type.startswith("Float") else tuple(reversed(_type.split(" of ")))
            # gets the text if the node is only text, extracts the text if the node contains tags (e.g. links)
            # then splits it to get all the types (if there is more than one)
            for _type in extract_all_text(cell).split(" or ")
        )

    def __add_entity(self, _type, name, params, endpoints, types):
        if _type == 0:
            endpoints.append((name, params))
        elif _type == 1:
            types[name] = (name, params)

    def parse(self, source: str):
        # avoid too much getattr machinery
        primitive_types = self.__primitive_types
        extract_text = self.__extract_text
        extract_all_text = self.__extract_all_text
        extract_type_list = self.__extract_type_list
        add_entity = self.__add_entity

        # parse the HTML source
        tree = etree.HTML(source)

        # get the version
        version = tree.xpath('//*[@id="dev_page_content"]/p[1]/strong')[0].text
        version = version.rsplit(" ", 1)[1]

        # node from where to start the parsing (the available methods section)
        starting_node = tree.xpath('//*[@id="dev_page_content"]/h3[text()="Getting updates"]')[0]

        # -1 is not valid, 0 is endpoint, 1 is type
        endpoints = []
        types = {}
        entity_type = -1

        # Main Type -> [Subtypes]
        aliases = {}

        # Subtype -> Main Type
        reversed_aliases = {}

        # loop variables
        name = None
        params = None

        for el in starting_node.itersiblings():
            # the telegram API page is very unfriendly to parse in some way:
            # you would expect each endpoint definition to be contained in a div or something but it's not, it's just a
            # stream of tags so we have to delimit a sequence of tags that makes sense
            if el.tag == "h4":
                # <h4> signals a new entity
                # put the current entity in the correct array, if applicable
                add_entity(entity_type, name, params, endpoints, types)

                # h4 tags are structured like <h4><a>link</a>ENDPOINT_NAME</a>
                # since we want that "ENDPOINT_NAME" part extracting text by XPath does that
                name = extract_text(el)
                if " " in name:
                    entity_type = -1
                    continue

                # if the name begins with a lower letter it's an endpoint, else it's a type (bool(True) = 1)
                entity_type = int(name[0].isupper())

                params = ()
            elif el.tag == "p":
                # get description
                pass
            elif el.tag == "h3":
                # <h3> signals an end of section: append the current entity
                add_entity(entity_type, name, params, endpoints, types)

                # avoid appending anything else later
                entity_type = -1
            elif el.tag == "ul":
                # entities with an "ul" are lists of names that should be used instead of it
                # insert everything inside the aliases dict
                if entity_type == -1:
                    continue

                # each <li> contains an <a>
                aliases[name] = [alias[0].text for alias in el.iterchildren()]
                reversed_aliases.update((alias[0].text, name) for alias in el.iterchildren())

                # don't append the type since we have it already
                entity_type = -1
            elif el.tag == "table":
                # get the <tbody>
                table = el[1]

                # normalize the entity into an array of tuples (name, type list, required ? True : False)
                if entity_type:
                    # entity_type = 1 --> type
                    # row[0] = name
                    # row[1] = type
                    # row[2] = description (begins with "Optional" if optional)
                    params = tuple(
                        # fmt: off
                        IndexedTuple((
                            row[0].text,
                            extract_type_list(row[1], extract_all_text),
                            not ((desc_cell := row[2]).text or extract_all_text(desc_cell)).startswith("Optional"),
                        ))
                        # fmt: on
                        for row in table.iterchildren()
                    )
                else:
                    # entity_type = 0 --> endpoint
                    # row[0] = name
                    # row[1] = type
                    # row[2] = Yes if required, Optional if optional
                    # row[3] = description
                    params = tuple(
                        # fmt: off
                        IndexedTuple((
                            row[0].text,
                            extract_type_list(row[1], extract_all_text),
                            row[2].text == "Yes"
                        ))
                        # fmt: on
                        for row in table.iterchildren()
                    )

        # append the last one
        add_entity(entity_type, name, params, endpoints, types)

        # actually used types (in correct order)
        used_types_deduped = []

        # get alias common params
        for alias, type_names in aliases.items():
            # build a set with the first type params
            _, first_type_params = types[type_names[0]]
            common_params = set(first_type_params)

            # keep only the intersection with each alias
            for type_name in type_names[1:]:
                _, params = types[type_name]
                common_params.intersection_update(params)

            used_types_deduped.append((alias, common_params))

            # remove common params from child classes
            for type_name in type_names:
                _, params = types[type_name]
                params = set(params)
                params.difference_update(common_params)
                types[type_name] = (type_name, params)

        # keep only the types that are actually used
        used_types = []

        used_type_names = set(
            base_type
            for endpoint in endpoints
            for param in endpoint[1]
            for _type in param[1]
            if (base_type := _type[0]) not in primitive_types
        )
        _used_type_names = set()

        used_type_names_arr = [used_type_names, _used_type_names]
        used_type_names_idx = 0

        # keep track of the types used by the types to process them later
        while used_type_names:
            for type_name in used_type_names:
                if (type_aliases := aliases.get(type_name, None)) :
                    # if the type has aliases we process it and add the aliases
                    _used_type_names.update(alias for alias in type_aliases)

                elif (_type := types.get(type_name, None)) :
                    # if we have found a definition add the definition and process other custom types (if there's any)
                    used_types.append(_type)
                    _used_type_names.update(
                        base_type
                        for param in _type[1]
                        for param_type in param[1]
                        if (base_type := param_type[0]) not in primitive_types
                    )

            # iterate over the remaining types
            used_type_names_idx = 1 - used_type_names_idx
            used_type_names.clear()
            used_type_names = used_type_names_arr[used_type_names_idx]
            _used_type_names = used_type_names_arr[1 - used_type_names_idx]

        # we insert each type an a list and deduplicate them later to keep them in the correct order
        used_types_deduped.extend(
            _type
            for _type in reversed(used_types)
            if not ((name := _type[0]) in used_type_names or used_type_names.add(name))
        )

        return version, endpoints, used_types_deduped, reversed_aliases



from json import dumps as json_dumps
from typing import List, Optional, Tuple, Union
from uuid import uuid4

from .generated import InputMedia, InputMediaPhoto, InputMediaVideo
from .multipart_encoder import MultipartEncoder
from .types import InlineKeyboardMarkup


def sendMediaGroup(
    chat_id: Union[int, str],
    media: List[Union[InputMediaPhoto, InputMediaVideo]],
    *,
    disable_notification: Optional[bool] = None,
    reply_to_message_id: Optional[int] = None
):
    files: List[Tuple[bytes, bytes, bytes, bytes]] = []

    params: dict = {"chat_id": chat_id}

    for _media in media:
        _media["media"]("media", files, _media, attach=uuid4().hex)  # type: ignore

    params["media"] = json_dumps(media, check_circular=False)

    if disable_notification is not None:
        params["disable_notification"] = disable_notification

    if reply_to_message_id is not None:
        params["reply_to_message_id"] = reply_to_message_id

    if files:
        headers = {}
        encoder = MultipartEncoder(files=files)
        headers["content-type"], body = encoder.encode()
        return "POST", "sendMediaGroup", headers, params, body
    else:
        return "GET", "sendMediaGroup", None, params, None


def editMessageMedia(
    media: InputMedia,
    *,
    chat_id: Optional[Union[int, str]] = None,
    message_id: Optional[int] = None,
    inline_message_id: Optional[str] = None,
    reply_markup: Optional[InlineKeyboardMarkup] = None
):
    files: List[Tuple[bytes, bytes, bytes, bytes]] = []

    params: dict = {}

    media["media"]("media", files, media, attach="media")  # type: ignore
    params["media"] = json_dumps(media, check_circular=False)

    if chat_id is not None:
        params["chat_id"] = chat_id

    if message_id is not None:
        params["message_id"] = message_id

    if inline_message_id is not None:
        params["inline_message_id"] = inline_message_id

    if reply_markup is not None:
        params["reply_markup"] = json_dumps(reply_markup.serialized)

    if files:
        headers = {}
        encoder = MultipartEncoder(files=files)
        headers["content-type"], body = encoder.encode()
        return "POST", "editMessageMedia", headers, params, body
    else:
        return "GET", "editMessageMedia", None, params, None



from asyncio import gather as asyncio_gather, get_event_loop
from inspect import currentframe
from json import loads as json_loads
from os import environ
from os.path import abspath, basename, dirname
from unittest import IsolatedAsyncioTestCase
from urllib.parse import urlencode

from telegen_definitions import (
    ForceReply,
    InlineKeyboardMarkup,
    InputFile,
    InputMediaPhoto,
    ReplyKeyboardRemove,
    ReplyKeyboardMarkup,
    editMessageMedia,
    sendDocument,
    sendMediaGroup,
    sendMessage,
    sendPhoto,
)

from ._http import HttpConnectionPool


token = environ.get("TELEGRAM_TOKEN")
chat_id = environ.get("TELEGRAM_CHAT_ID")


def InputFile_from_path(path):
    with open(path, "rb") as f:
        content = f.read()
    return InputFile(basename(path), body=content)


current_dir = dirname(abspath(__file__))
document = InputFile_from_path(f"{current_dir}/document.pdf")
photo = InputFile_from_path(f"{current_dir}/photo.jpg")
photo2 = InputFile_from_path(f"{current_dir}/photo2.jpg")
photo_url = InputFile("https://placekitten.com/200/200")


class ApiCallsCase(IsolatedAsyncioTestCase):
    async def test_calls(self):
        pool = HttpConnectionPool("api.telegram.org", loop=get_event_loop(), size=5)

        async def telegram_req(req):
            # "GET", "sendMessage", None, params, None
            method, endpoint, headers, params, body = req

            url = f"bot{token}/{endpoint}"
            if params:
                generator = [(k, v) for k, v in params.items() if v]
                url = url + "?" + urlencode(generator)

            try:
                res = await pool.request(url, method=method, headers=headers, body=body)
            except Exception as e:
                print(e)
                print(method, endpoint, headers, params, body)
                raise e

            res = json_loads(res.decode())
            assert res["ok"] == True

            return res

        async def test_sendMessage():
            req = sendMessage(chat_id, "Test!")
            res = await telegram_req(req)

        async def test_sendDocument():
            req = sendDocument(chat_id, document, thumb=photo)
            res = await telegram_req(req)

        async def test_sendPhotoUrl():
            req = sendPhoto(chat_id, photo_url)
            res = await telegram_req(req)

        async def test_editMessageMedia():
            req = sendPhoto(chat_id, photo_url)
            res = await telegram_req(req)

            new_photo: InputMediaPhoto = {"media": photo2, "type": "photo"}
            req = editMessageMedia(new_photo, chat_id=chat_id, message_id=res["result"]["message_id"])
            res = await telegram_req(req)

        async def test_editMessageMediaUrl():
            req = sendPhoto(chat_id, photo)
            res = await telegram_req(req)

            new_photo: InputMediaPhoto = {"media": photo_url, "type": "photo"}
            req = editMessageMedia(new_photo, chat_id=chat_id, message_id=res["result"]["message_id"])
            res = await telegram_req(req)

        async def test_sendMediaGroup():
            photos: List[InputMediaPhoto] = [
                {"media": photo, "type": "photo"},
                {"media": photo2, "type": "photo"},
                {"media": photo_url, "type": "photo"},
            ]

            req = sendMediaGroup(chat_id, photos)
            res = await telegram_req(req)

        async def test_InlineKeyboardMarkup():
            keyboard = InlineKeyboardMarkup().callback_data("test", "test")
            req = sendMessage(chat_id, "Test", reply_markup=keyboard)
            res = await telegram_req(req)

        async def test_ReplyKeyboard():
            keyboard = ReplyKeyboardMarkup().add_button("Test").add_row().add_button("Test2").add_button("Test3")
            req = sendMessage(chat_id, "Test", reply_markup=keyboard)
            res = await telegram_req(req)

            req = sendMessage(chat_id, "Test", reply_markup=ReplyKeyboardRemove())
            res = await telegram_req(req)

        async def test_ForceReply():
            req = sendMessage(chat_id, "Test", reply_markup=ForceReply().selective())
            res = await telegram_req(req)

        frame = currentframe()
        test_fns = [val() for name, val in frame.f_locals.items() if name.startswith("test_")]

        await asyncio_gather(*test_fns)

        pool.stop()



from asyncio import Protocol, Queue as AsyncQueue


class TCP(Protocol):
    __slots__ = ["transport", "future"]

    def connection_made(self, transport):
        self.transport = transport

    def connection_lost(self, exc):
        if exc:
            self.future.set_exception(exc)

    def data_received(self, data):
        body = data.split(b"\r\n\r\n", 1)[1]
        self.future.set_result(body)

    def eof_received(self):
        return False


class HttpConnectionPool:
    def __init__(self, host, *, loop, size):
        self.host = host
        self.loop = loop

        self.queue = AsyncQueue()
        self.tasks = [loop.create_task(self.open()) for _ in range(size)]

    async def open(self):
        transport = None

        queue = self.queue
        loop = self.loop
        host = self.host

        while True:
            (future, req) = await queue.get()

            if transport is None:
                (transport, protocol) = await loop.create_connection(lambda: TCP(), host, 443, ssl=True)

            protocol.future = future
            transport.write(req)
            await future

            if queue.empty():
                transport.close()
                transport = None

    def stop(self):
        for task in self.tasks:
            task.cancel()

    def request(self, path, method="GET", headers=None, body=None):
        # fmt: off
        req = (
            f"{method} /{path} HTTP/1.1\r\n"
            f"host: {self.host}\r\n"
        )
        # fmt: on

        if headers:
            req += "".join(f"{k}: {v}\r\n" for k, v in headers.items())

        if body:
            req += f"content-length: {len(body)}\r\n\r\n"
            req = req.encode()
            req += body
        else:
            req += "\r\n"
            req = req.encode()

        future = self.loop.create_future()
        self.queue.put_nowait((future, req))
        return future



import setuptools

with open("version.txt", "r") as f:
    version = f.read()

with open("README.rst", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="telegen-definitions",
    version=version,
    author="Alessandro Cerruti",
    author_email="strychnide@protonmail.com",
    description="Autogenerated telegen definitions",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://gitlab.com/strychnide/telegen",
    project_urls={
        "Documentation": "https://strychnide.gitlab.io/telegen",
        "Issues": "https://gitlab.com/strychnide/telegen/-/issues",
        "Source": "https://gitlab.com/strychnide/telegen",
    },
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Communications :: Chat",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    packages=["telegen_definitions"],
    python_requires=">=3.6",
)



from argparse import ArgumentParser, FileType
from os import system
from os.path import dirname, realpath
from shutil import copytree, rmtree
from urllib.request import urlopen

from .parser import ApiParser
from .generators.python import PythonApiGenerator


def main():
    # get the arguments
    parser = ArgumentParser(
        prog="telegen", description="A CLI tool that generates code by parsing the Telegram Bot API page"
    )
    parser.add_argument("language", help="Output language")
    parser.add_argument(
        "src", help="Source file (will be fetched if not passed)", nargs="?", type=FileType("r"), default=None,
    )
    parser.add_argument(
        "--overwrite", "-o", help="Delete conflicting files before emitting (default: exit)", action="store_true"
    )
    args = parser.parse_args()

    # variables
    module_dir = dirname(realpath(__file__))
    language = args.language
    output_dir = f"telegen_{language}"

    # delete the output directory if the flag is set and prepare a new one
    if args.overwrite:
        rmtree(output_dir, ignore_errors=True)

    copytree(f"{module_dir}/includes/{language}", output_dir)

    # get the source from the given file or fetch it from the URL
    src = (args.src or urlopen("https://core.telegram.org/bots/api")).read()

    # generate the code
    parser = ApiParser()
    version, endpoints, types, reversed_aliases = parser.parse(src)

    languages = {"python": PythonApiGenerator}

    generator = languages[language]()
    generated = generator.generate(version, endpoints, types, reversed_aliases)

    # write the files
    for filename, content in generated.items():
        with open(f"{output_dir}/{filename}", "w+") as f:
            f.write(content)



import setuptools

with open("README.rst", "r") as f:
    long_description = f.read()


pkg_info = {}
with open("telegen/__init__.py", "r") as f:
    module_info = exec(f.read(), pkg_info)


setuptools.setup(
    name="telegen",
    version=pkg_info["__version__"],
    author="Alessandro Cerruti",
    author_email="strychnide@protonmail.com",
    description="A CLI tool that generates code by parsing the Telegram Bot API page",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://gitlab.com/strychnide/telegen",
    project_urls={
        "Documentation": "https://strychnide.gitlab.io/telegen",
        "Issues": "https://gitlab.com/strychnide/telegen/-/issues",
        "Source": "https://gitlab.com/strychnide/telegen",
    },
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Code Generators",
    ],
    packages=setuptools.find_packages(include=["telegen", "telegen.*"]),
    include_package_data=True,
    python_requires=">=3.8",
    install_requires=["black", "lxml"],
    extras_require={"doc": ["sphinx", "sphinx_rtd_theme"],},
    entry_points={"console_scripts": ["telegen=telegen.cli:main"]},
)



from black import FileMode, format_file_contents

from telegen import __version__


class PythonApiGenerator:
    __slots__ = [
        "__space",
        "__custom_types",
        "__custom_endpoints",
        "__type_to_type_map",
        "__name_to_type_map",
        "__name_to_value_map",
        "__imports_str",
    ]

    def __init__(self):
        self.__space = " " * 4
        SPACE = self.__space

        # types that must be implemented by hand
        self.__custom_types = {
            # InputFile and related
            "InputFile",
            # ReplyMarkup
            "InlineKeyboardMarkup",
            "ReplyKeyboardMarkup",
            "ReplyKeyboardRemove",
            "ForceReply",
            # consequences of ReplyMarkup
            "KeyboardButton",
            "InlineKeyboardButton",
        }

        self.__custom_endpoints = {"sendMediaGroup", "editMessageMedia"}

        # maps a Telegram APi param *type* into a Python type
        self.__type_to_type_map = {
            "String": "str",
            "Integer": "int",
            "Float": "float",
            "Boolean": "bool",
        }

        # maps a Telegram API param *name* to a Python type
        self.__name_to_type_map = {
            # InputFile
            "animation": "InputFile",
            "audio": "InputFile",
            "certificate": "InputFile",
            "document": "InputFile",
            "photo": "InputFile",
            "png_sticker": "InputFile",
            "sticker": "InputFile",
            "tgs_sticker": "InputFile",
            "thumb": "InputFile",
            "video": "InputFile",
            "video_note": "InputFile",
            "voice": "InputFile",
            # the media field is announced as String, good job telegram monkeys!
            "media": "InputFile",
            # other types
            "parse_mode": "ParseMode",
        }

        # maps a Telegram API param name to his value (e.g. a property on an object)
        self.__name_to_value_map = {
            # InputFile
            "animation": 'animation("animation", files)',
            "audio": 'audio("audio", files)',
            "certificate": 'certificate("certificate", files)',
            "document": 'document("document", files)',
            "photo": 'photo("photo", files)',
            "png_sticker": 'png_sticker("png_sticker", files)',
            "sticker": 'sticker("sticker", files)',
            "tgs_sticker": 'tgs_sticker("tgs_sticker", files)',
            "thumb": 'thumb("thumb", files, attach="thumb")',
            "video": 'video("video", files)',
            "video_note": 'video_note("video_note", files)',
            "voice": 'voice("voice", files)',
            # other types
            "parse_mode": "parse_mode.name",
            "reply_markup": "json_dumps(reply_markup.serialized, check_circular=False)",
        }

        # added to the beginning of the generated file
        self.__imports_str = (
            f"from json import dumps as json_dumps\n"
            f"from typing import List, Optional, Tuple, Union\n\n"
            f"try:\n"
            f"{SPACE}from typing import TypedDict\n"
            f"except:\n"
            f"{SPACE}try:\n"
            f"{SPACE}{SPACE}from mypy_extensions import TypedDict\n"
            f"{SPACE}except:\n"
            f"{SPACE}{SPACE}TypedDict = dict\n\n"
            f"from .multipart_encoder import MultipartEncoder\n"
            f"from .types import *\n\n"
            f"\n"
        )

    def __build_param_definition(self, name, types, type_to_type_map, name_to_type_map, name_to_value_map):
        if len(types) > 1:
            # if the types array has length > 1 it's a Union
            type_def = name_to_type_map.get(
                # check if the name is associated to a type
                name,
                # or compose the type using the definitions:
                # if the single type has length > 1 it's a List[Type]
                # e.g. for length = 3 the type is a List[List[Type]]
                # for each type use the corresponding Python type if there's one (e.g. String --> str)
                # or use the name itself if not
                (
                    "Union["
                    + ", ".join(
                        f"List[" * (size := len(_type) - 1)
                        + type_to_type_map.get((base_type := _type[0]), base_type)
                        + "]" * size
                        for _type in types
                    )
                    + "]"
                ),
            )
        else:
            _type = types[0]
            base_type = _type[0]
            type_def = name_to_type_map.get(
                name, f"List[" * (size := len(_type) - 1) + type_to_type_map.get(base_type, base_type) + "]" * size
            )

        # derive the actual value from the value of the parameter
        # print(types[])
        value = name_to_value_map.get(
            name, f"json_dumps({name}, check_circular=False)" if types[0][0] not in type_to_type_map else name,
        )

        return type_def, value

    def generate(self, version, endpoints, types, reversed_aliases):
        # avoid too much getattr machinery
        SPACE = self.__space
        custom_types = self.__custom_types
        custom_endpoints = self.__custom_endpoints
        type_to_type_map = self.__type_to_type_map
        name_to_type_map = self.__name_to_type_map
        name_to_value_map = self.__name_to_value_map
        imports = self.__imports_str
        build_param_definition = self.__build_param_definition

        endpoint_defs = []
        for endpoint in endpoints:
            _name, params = endpoint

            if _name in custom_endpoints:
                continue

            required_params = []
            optional_params = []
            params_decl = []
            after_params = []
            body = []

            input_file_pre = ""
            input_file_post = ""

            input_media = ""

            for param in params:
                name, _types, required = param

                type_def, value = build_param_definition(
                    name, _types, type_to_type_map, name_to_type_map, name_to_value_map
                )

                # ugly special cases
                # see the descriptions in the constructor and the InputFile definition and you'll get how this works
                if type_def == "InputFile":
                    # if the method has an InputFile or InputMedia there will be some params that fill files if the media is not an url or id
                    input_file_pre = f"{SPACE}files: List[Tuple[bytes, bytes, bytes, bytes]] = []\n\n"

                    # if the media are filled we must send a POST request, else we return the usual GET
                    input_file_post = (
                        f"{SPACE}if files:\n"
                        f"{SPACE}{SPACE}headers = {{}}\n"
                        f"{SPACE}{SPACE}encoder = MultipartEncoder(files=files)\n"
                        f'{SPACE}{SPACE}headers["content-type"], body = encoder.encode()\n'
                        f'{SPACE}{SPACE}return "POST", "{_name}", headers, params, body\n'
                        f"{SPACE}else:\n{SPACE}"
                    )

                    attach = '"thumb"' if name == "thumb" else "None"

                    if required:
                        required_params.append(f"{name}: {type_def}")
                        after_params.append(f'{SPACE}{name}("{name}", files, params, attach={attach})\n\n')
                    else:
                        optional_params.append(f"{name}: Optional[{type_def}] = None")
                        after_params.append(
                            f"{SPACE}if {name} is not None:\n"
                            f'{SPACE}{SPACE}{SPACE}{name}("{name}", files, params, attach={attach})\n\n'
                        )
                else:
                    if required:
                        # if the type is required there's no default
                        required_params.append(f"{name}: {type_def}")

                        # add the param to the dict initialization
                        params_decl.append(f'{SPACE}{SPACE}"{name}": {value}')
                    else:
                        # if the type is optional his default value is None
                        optional_params.append(f"{name}: Optional[{type_def}] = None")

                        # add the param to the dict if there's no default
                        # don't format the following lines for clarity
                        # fmt: off
                        body.append(
                            f"{SPACE}if {name} is not None:\n"
                            f'{SPACE}{SPACE}params["{name}"] = {value}\n\n'
                        )
                        # fmt: on

            if required_params:
                # if there are required parameters we build the dict and init it with the values
                params_return_var = "params"
                params_decl = f"{SPACE}params: dict = {{\n" + f",\n".join(params_decl) + f"\n{SPACE}}}\n\n"

                # the only defference from the "only optional" case is that here we put a "," after the required params
                # and before the optional params (if there are any)
                required_params = ", ".join(required_params)
                optional_params = (", *, " + ", ".join(optional_params)) if optional_params else ""
            elif optional_params:
                # if there are only optional parameters we create the dict but return None if the dict is not good
                params_return_var = "params or None"
                params_decl = f"{SPACE}params: dict = {{}}\n\n"

                required_params = ""
                optional_params = "*, " + ", ".join(optional_params)
            else:
                # if there are no params we return None and don't even create the dict
                params_return_var = "None"
                params_decl = ""
                required_params = ""
                optional_params = ""

            body = "".join(body)
            after_params = "".join(after_params)

            endpoint_defs.append(
                f"def {_name}({required_params}{optional_params}) -> EndpointCall:\n"
                f"{input_file_pre}"
                f"{input_media}"
                f"{params_decl}"
                f"{after_params}"
                f"{body}"
                f"{input_file_post}"
                f'{SPACE}return "GET", "{_name}", None, {params_return_var}, None\n'
            )
            # print(endpoint_defs[-1])

        type_defs = []
        for _type in types:
            _name, params = _type

            if _name in custom_types:
                continue

            required_params = []
            optional_params = []

            for param in params:
                name, _types, required = param

                type_def, value = build_param_definition(
                    name, _types, type_to_type_map, name_to_type_map, name_to_value_map
                )

                if required:
                    required_params.append(f"{SPACE}{name}: {type_def}")
                else:
                    optional_params.append(f"{SPACE}{name}: Optional[{type_def}]")

            if required_params:
                required_params = "\n".join(required_params)
                optional_params = ("\n" + "\n".join(optional_params)) if optional_params else ""
            else:
                required_params = ""
                optional_params = "\n".join(optional_params) or f"{SPACE}pass"

            base_class = reversed_aliases.get(_name, "TypedDict")

            # fmt: off
            type_defs.append(
                f"class {_name}({base_class}, total=False):\n"
                f"{required_params}"
                f"{optional_params}\n"
            )
            # fmt: on

        code = imports + "".join(type_defs) + "".join(endpoint_defs)
        code = format_file_contents(code, fast=False, mode=FileMode())

        return {"telegen_definitions/generated.py": code, "version.txt": f"{version}.{__version__}"}
